searchData={"items":[{"type":"module","title":"Hammer.Redis","doc":"This backend uses the [Redix](https://hex.pm/packages/redix) library to connect to Redis.\n\n    defmodule MyApp.RateLimit do\n      # the default prefix is \"MyApp.RateLimit:\"\n      # the default timeout is :infinity\n      use Hammer, backend: Hammer.Redis, prefix: \"MyApp.RateLimit:\", timeout: :infinity\n    end\n\n    MyApp.RateLimit.start_link(url: \"redis://localhost:6379\")\n\n    # increment and timeout arguments are optional\n    # by default increment is 1 and timeout is as defined in the module\n    {:allow, _count} = MyApp.RateLimit.hit(key, scale, limit)\n    {:allow, _count} = MyApp.RateLimit.hit(key, scale, limit, _increment = 1, _timeout = :infinity)","ref":"Hammer.Redis.html"},{"type":"type","title":"Hammer.Redis.redis_option/0","doc":"","ref":"Hammer.Redis.html#t:redis_option/0"},{"type":"type","title":"Hammer.Redis.redis_options/0","doc":"","ref":"Hammer.Redis.html#t:redis_options/0"},{"type":"module","title":"Hammer.Redis.FixWindow","doc":"This module implements the Fix Window algorithm.\n\nThe fixed window algorithm works by dividing time into fixed intervals or \"windows\"\nof a specified duration (scale). Each window tracks request counts independently.\n\nFor example, with a 60 second window:\n- Window 1: 0-60 seconds\n- Window 2: 60-120 seconds\n- And so on...\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate which window it belongs to based on current time\n   - Increment the counter for that window\n   - Store expiration time as end of window\n2. To check if rate limit is exceeded:\n   - If count <= limit: allow request\n   - If count > limit: deny and return time until window expires\n3. Old windows are automatically cleaned up after expiration\n\nThis provides simple rate limiting but has edge cases where a burst of requests\nspanning a window boundary could allow up to 2x the limit in a short period.\nFor more precise limiting, consider using the sliding window algorithm instead.\n\nThe fixed window algorithm is a good choice when:\n\n- You need simple, predictable rate limiting with clear time boundaries\n- The exact precision of the rate limit is not critical\n- You want efficient implementation with minimal storage overhead\n- Your use case can tolerate potential bursts at window boundaries\n\nCommon use cases include:\n\n- Basic API rate limiting where occasional bursts are acceptable\n- Protecting backend services from excessive load\n- Implementing fair usage policies\n- Scenarios where clear time-based quotas are desired (e.g. \"100 requests per minute\")\n\nThe main tradeoff is that requests near window boundaries can allow up to 2x the\nintended limit in a short period. For example with a limit of 100 per minute:\n- 100 requests at 11:59:59\n- Another 100 requests at 12:00:01\n\nThis results in 200 requests in 2 seconds, while still being within limits.\nIf this behavior is problematic, consider using the sliding window algorithm instead.","ref":"Hammer.Redis.FixWindow.html"},{"type":"module","title":"Hammer.Redis.LeakyBucket","doc":"This module implements the Leaky Bucket algorithm.\n\nThe leaky bucket algorithm works by modeling a bucket that:\n- Fills up with requests at the input rate\n- \"Leaks\" requests at a constant rate\n- Has a maximum capacity (the bucket size)\n\nFor example, with a leak rate of 10 requests/second and bucket size of 100:\n- Requests add to the bucket's current level\n- The bucket leaks 10 requests per second steadily\n- If bucket reaches capacity (100), new requests are denied\n- Once bucket level drops, new requests are allowed again\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate how much has leaked since last request\n   - Subtract leaked amount from current bucket level\n   - Try to add new request to bucket\n   - Store new bucket level and timestamp\n2. To check if rate limit is exceeded:\n   - If new bucket level <= capacity: allow request\n   - If new bucket level > capacity: deny and return time until enough leaks\n3. Old entries are automatically cleaned up after expiration\n\nThis provides smooth rate limiting with ability to handle bursts up to bucket size.\nThe leaky bucket is a good choice when:\n\n- You need to enforce a constant processing rate\n- Want to allow temporary bursts within bucket capacity\n- Need to smooth out traffic spikes\n- Want to prevent resource exhaustion\n\nCommon use cases include:\n\n- API rate limiting needing consistent throughput\n- Network traffic shaping\n- Service protection from sudden load spikes\n- Queue processing rate control\n- Scenarios needing both burst tolerance and steady-state limits\n\nThe main advantages are:\n- Smooth, predictable output rate\n- Configurable burst tolerance\n- Natural queueing behavior\n\nThe tradeoffs are:\n- More complex implementation than fixed windows\n- Need to track last request time and current bucket level\n- May need tuning of bucket size and leak rate parameters\n\nFor example, with 100 requests/sec limit and 500 bucket size:\n- Can handle bursts of up to 500 requests\n- But long-term average rate won't exceed 100/sec\n- Provides smoother traffic than fixed windows","ref":"Hammer.Redis.LeakyBucket.html"},{"type":"function","title":"Hammer.Redis.LeakyBucket.get/4","doc":"Returns the current level of the bucket for a given key.","ref":"Hammer.Redis.LeakyBucket.html#get/4"},{"type":"function","title":"Hammer.Redis.LeakyBucket.hit/7","doc":"Checks if a key is allowed to perform an action, and increment the counter by the given amount.","ref":"Hammer.Redis.LeakyBucket.html#hit/7"},{"type":"module","title":"Hammer.Redis.TokenBucket","doc":"This module implements the Token Bucket algorithm.\nThe token bucket algorithm works by modeling a bucket that:\n- Fills with tokens at a constant rate (the refill rate)\n- Has a maximum capacity of tokens (the bucket size)\n- Each request consumes one or more tokens\n- If there are enough tokens, the request is allowed\n- If not enough tokens, the request is denied\n\nFor example, with a refill rate of 10 tokens/second and bucket size of 100:\n- Tokens are added at 10 per second up to max of 100\n- Each request needs tokens to proceed\n- If bucket has enough tokens, request allowed and tokens consumed\n- If not enough tokens, request denied until bucket refills\n\nThe algorithm:\n1. When a request comes in, we:\n   - Calculate tokens added since last request based on time elapsed\n   - Add new tokens to bucket (up to max capacity)\n   - Try to consume tokens for the request\n   - Store new token count and timestamp\n2. To check if rate limit is exceeded:\n   - If enough tokens: allow request and consume tokens\n   - If not enough: deny and return time until enough tokens refill\n3. Old entries are automatically cleaned up after expiration\n\nThis provides smooth rate limiting with ability to handle bursts up to bucket size.\nThe token bucket is a good choice when:\n\n- You need to allow temporary bursts of traffic\n- Want to enforce an average rate limit\n- Need to support different costs for different operations\n- Want to avoid the sharp edges of fixed windows\n\nCommon use cases include:\n\n- API rate limiting with burst tolerance\n- Network traffic shaping\n- Resource allocation control\n- Gaming systems with \"energy\" mechanics\n- Scenarios needing flexible rate limits\n\nThe main advantages are:\n- Natural handling of bursts\n- Flexible token costs for different operations\n- Smooth rate limiting behavior\n- Simple to reason about\n\nThe tradeoffs are:\n- Need to track token count and last update time\n- May need tuning of bucket size and refill rate\n- More complex than fixed windows\n\nFor example with 100 tokens/minute limit and 500 bucket size:\n- Can handle bursts using saved up tokens\n- Automatically smooths out over time\n- Different operations can cost different amounts\n- More flexible than fixed request counts","ref":"Hammer.Redis.TokenBucket.html"},{"type":"function","title":"Hammer.Redis.TokenBucket.get/4","doc":"Returns the current level of the bucket for a given key.","ref":"Hammer.Redis.TokenBucket.html#get/4"},{"type":"extras","title":"Hammer.Redis","doc":"# Hammer.Redis\n\n[![Build Status](https://github.com/ExHammer/hammer-backend-redis/actions/workflows/ci.yml/badge.svg)](https://github.com/ExHammer/hammer-backend-redis/actions/workflows/ci.yml)\n[![Hex.pm](https://img.shields.io/hexpm/v/hammer_backend_redis.svg)](https://hex.pm/packages/hammer_backend_redis)\n[![Documentation](https://img.shields.io/badge/documentation-gray)](https://hexdocs.pm/hammer_backend_redis)\n[![Total Download](https://img.shields.io/hexpm/dt/hammer_backend_redis.svg)](https://hex.pm/packages/hammer_backend_redis)\n[![License](https://img.shields.io/hexpm/l/hammer_backend_redis.svg)](https://github.com/ExHammer/hammer-backend-redis/blob/master/LICENSE.md)\n\nA Redis backend for the [Hammer](https://github.com/ExHammer/hammer) rate-limiter.\n\nThis backend is a thin [Redix](https://hex.pm/packages/redix) wrapper. A single connection is used per rate-limiter. It should be enough for most use-cases since packets for rate limiting requests are short (i.e. no head of line blocking) and Redis is OK with [pipelining](https://redis.io/learn/operate/redis-at-scale/talking-to-redis/client-performance-improvements#pipelining) (i.e. we don't block awaiting replies). Consider benchmarking before introducing more connections since TCP performance might be unintuitive. For possible pooling approaches, see Redix docs on [pooling](https://hexdocs.pm/redix/real-world-usage.html#name-based-pool) and also [PartitionSupervisor.](https://hexdocs.pm/elixir/1.17.3/PartitionSupervisor.html) Do not use poolboy or db_connection-like pools since they practically disable pipelining which leads to worse connection utilisation and worse performance.\n\nThe algorithm we are using is the first method described (called \"bucketing\") in [Rate Limiting with Redis](https://youtu.be/CRGPbCbRTHA?t=753).\nIn other sources it's sometimes called a \"fixed window counter\".\n\n**TODO:** document ttl issues if servers are misconfigured","ref":"readme.html"},{"type":"extras","title":"Installation - Hammer.Redis","doc":"Hammer-backend-redis\nis [available in Hex](https://hex.pm/packages/hammer_backend_redis), the package\ncan be installed by adding `hammer_backend_redis` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:hammer_backend_redis, \"~> 7.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Usage - Hammer.Redis","doc":"Define the rate limiter:\n\n```elixir\ndefmodule MyApp.RateLimit do\n  use Hammer, backend: Hammer.Redis\nend\n```\n\nAnd add it to your app's supervision tree:\n\n```elixir\nchildren = [\n  {MyApp.RateLimit, url: \"redis://localhost:6379\"}\n]\n```\n\nAnd that's it, calls to `MyApp.RateLimit.hit/3` and so on will use Redis to store\nthe rate-limit counters. See the [documentation](https://hexdocs.pm/hammer_backend_redis/Hammer.Redis.html) for more details.","ref":"readme.html#usage"},{"type":"extras","title":"Run tests locally - Hammer.Redis","doc":"You need a running Redis instance. One can be started locally using `docker compose up -d redis`.\nSee the [compose.yml](./compose.yml) for more details.","ref":"readme.html#run-tests-locally"},{"type":"extras","title":"Getting Help - Hammer.Redis","doc":"If you're having trouble, open an issue on this repo.","ref":"readme.html#getting-help"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"7.0.0-rc.1 (2024-12-18) - Changelog","doc":"","ref":"changelog.html#7-0-0-rc-1-2024-12-18"},{"type":"extras","title":"Changed - Changelog","doc":"- Added `:algorithm` option to the backend with support for:\n  - `:fix_window` (default) - Fixed time window rate limiting\n  - `:leaky_bucket` - Constant rate limiting with burst capacity\n  - `:token_bucket` - Token-based rate limiting with burst capacity\n- Add benchmarks file and run them with `bench`","ref":"changelog.html#changed"},{"type":"extras","title":"7.0.0-rc.0 (2024-12-06) - Changelog","doc":"","ref":"changelog.html#7-0-0-rc-0-2024-12-06"},{"type":"extras","title":"Changed - Changelog","doc":"- Conform to new Hammer API\n- Remove Poolboy as it introduces unnecessary blocking.","ref":"changelog.html#changed"},{"type":"extras","title":"6.2.0 (2024-12-04) - Changelog","doc":"","ref":"changelog.html#6-2-0-2024-12-04"},{"type":"extras","title":"Changed - Changelog","doc":"- Package updates\n- Add config to customize the redis prefix\n- Deprecate Elixir 1.12 as this are no longer supported","ref":"changelog.html#changed"},{"type":"extras","title":"6.1.2 (2022-11-11) - Changelog","doc":"","ref":"changelog.html#6-1-2-2022-11-11"},{"type":"extras","title":"Changed - Changelog","doc":"- Applied credo suggestions\n- Update dependencies","ref":"changelog.html#changed"},{"type":"extras","title":"6.1.1 (2022-11-11) - Changelog","doc":"","ref":"changelog.html#6-1-1-2022-11-11"},{"type":"extras","title":"Changed - Changelog","doc":"- package update and ownership transferred","ref":"changelog.html#changed"},{"type":"extras","title":"6.1.0 (2019-09-03) - Changelog","doc":"","ref":"changelog.html#6-1-0-2019-09-03"},{"type":"extras","title":"Changed - Changelog","doc":"- Return actual count upon bucket creation (thanks to @davelively14, https://github.com/ExHammer/hammer-backend-redis/pull/16)","ref":"changelog.html#changed"},{"type":"extras","title":"6.0.1 (2019-07-13) - Changelog","doc":"","ref":"changelog.html#6-0-1-2019-07-13"},{"type":"extras","title":"Added - Changelog","doc":"- Accept an optional `redis_url` option","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- Updated dependencies in test environment (thanks to @ono, https://github.com/ExHammer/hammer-backend-redis/pull/14)","ref":"changelog.html#changed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed a crash in `delete_buckets` (thanks to @ono, https://github.com/ExHammer/hammer-backend-redis/pull/15)","ref":"changelog.html#fixed"},{"type":"extras","title":"6.0.0 (2018-10-13) - Changelog","doc":"","ref":"changelog.html#6-0-0-2018-10-13"},{"type":"extras","title":"Changed - Changelog","doc":"- Raise an error if `expiry_ms` is not configured explicitly\n- Update the `redix` dependency to `~> 0.8`","ref":"changelog.html#changed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Actually honor `:redis_config`, as is claimed in the documentation","ref":"changelog.html#fixed"},{"type":"extras","title":"5.0.0 (2018-10-13) - Changelog","doc":"","ref":"changelog.html#5-0-0-2018-10-13"},{"type":"extras","title":"Changed - Changelog","doc":"- Update to the new custom-increment api","ref":"changelog.html#changed"},{"type":"extras","title":"4.0.3 (2018-05-08) - Changelog","doc":"","ref":"changelog.html#4-0-3-2018-05-08"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fix a rare crash, again related to race-conditions\n  (https://github.com/ExHammer/hammer-backend-redis/issues/11#issuecomment-387202359)","ref":"changelog.html#fixed"},{"type":"extras","title":"4.0.2 (2018-04-27) - Changelog","doc":"","ref":"changelog.html#4-0-2-2018-04-27"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed race-condition, really this time\n  (https://github.com/ExHammer/hammer-backend-redis/issues/11)","ref":"changelog.html#fixed"},{"type":"extras","title":"4.0.1 (2018-04-24) - Changelog","doc":"","ref":"changelog.html#4-0-1-2018-04-24"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed a race-condition that could cause crashes\n  (https://github.com/ExHammer/hammer-backend-redis/issues/11)","ref":"changelog.html#fixed"},{"type":"extras","title":"4.0.0 (2018-04-23) - Changelog","doc":"","ref":"changelog.html#4-0-0-2018-04-23"},{"type":"extras","title":"Changed - Changelog","doc":"- Update to `Hammer 4.0`","ref":"changelog.html#changed"},{"type":"extras","title":"3.0.0 (2018-02-20) - Changelog","doc":"","ref":"changelog.html#3-0-0-2018-02-20"},{"type":"extras","title":"Changed - Changelog","doc":"- Require elixir >= 1.6","ref":"changelog.html#changed"},{"type":"extras","title":"2.0.0 (2017-09-24) - Changelog","doc":"","ref":"changelog.html#2-0-0-2017-09-24"},{"type":"extras","title":"Changed - Changelog","doc":"- Updated to new Hammer API","ref":"changelog.html#changed"},{"type":"extras","title":"1.0.0 (2017-08-27) - Changelog","doc":"","ref":"changelog.html#1-0-0-2017-08-27"},{"type":"extras","title":"Changed - Changelog","doc":"- `hammer_backend_redis` now explicitly depends on `hammer`\n- Implement the `Hammer.Backend` behaviour\n- Alias `redix_config` to `redis_config` in the config list, for convenience","ref":"changelog.html#changed"},{"type":"extras","title":"0.1.0 (2017-07-31) - Changelog","doc":"Initial release.","ref":"changelog.html#0-1-0-2017-07-31"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.35.1"}}